# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime
from enum import Enum
import logging

from fastcore.basics import AttrDict
from fastcore.net import HTTP4xxClientError
from ghapi.page import date2gh

GITHUB_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S%z'


def gh2date(dtstr):
    return datetime.strptime(dtstr, GITHUB_DATE_FORMAT)


class RepositoryItemType(Enum):
    """The types of items we are dealing with from GitHub."""

    ISSUES = 1
    COMMENTS = 2
    TEAMS = 3
    LABELS = 4
    EVENTS = 5
    COMMITS = 6
    RELEASES = 7


class RepositoryItem(AttrDict):
    """An item from a GitHub repository.
    
    This class extends the AttrDict class used by GhApi to provide
    helper methods to manipulate the items.
    """

    @property
    def creation_time(self):
        return datetime.strptime(self.created_at, GITHUB_DATE_FORMAT)

    @property
    def update_time(self):
        return datetime.strptime(self.updated_at, GITHUB_DATE_FORMAT)

    def _filter_time(self, time, after=None, before=None):
        return (not after or time > after) and (not before or time < before)

    def created(self, after=None, before=None):
        """Indicates whether the item was created in a given time span."""

        return self._filter_time(self.creation_time, after, before)

    def updated(self, after=None, before=None):
        """Indicates whether the item was updated in a given time span."""

        return self._filter_time(self.update_time, after, before)


class IssueItem(RepositoryItem):
    """A GitHub issue.
    
    This can be either an actual "issue" or a "pull requests".
    """

    @property
    def close_time(self):
        if not self.closed_at:
            return None
        return datetime.strptime(self.closed_at, GITHUB_DATE_FORMAT)

    def closed(self, after=None, before=None):
        """Indicates whether the issue was closed in a given time span."""

        if not self.closed_at:
            return False
        return self._filter_time(self.close_time, after, before)


class CommitItem(RepositoryItem):
    """A git commit."""

    @property
    def creation_time(self):
        return datetime.strptime(self.commit.author.date, GITHUB_DATE_FORMAT)


class RepositoryProvider(object):
    """Provides access to the data from a GitHub repository."""

    def get_data(self, item_type, since=None):
        """Gets a specific type of data from the repository.
        
        :param item_type: the type of data to fetch
        :param since: only fetch data from after that timestamp
        :return: an array of AttrDict objects
        """

        pass

    @property
    def issues(self):
        return [i for i in self.get_data(RepositoryItemType.ISSUES)
                if not hasattr(i, 'pull_request')]

    @property
    def pull_requests(self):
        return [i for i in self.get_data(RepositoryItemType.ISSUES)
                if hasattr(i, 'pull_request')]

    @property
    def comments(self):
        return self.get_data(RepositoryItemType.COMMENTS)

    @property
    def teams(self):
        return self.get_data(RepositoryItemType.TEAMS)

    @property
    def labels(self):
        return self.get_data(RepositoryItemType.LABELS)

    @property
    def events(self):
        return self.get_data(RepositoryItemType.EVENTS)

    @property
    def commits(self):
        return self.get_data(RepositoryItemType.COMMITS)

    @property
    def releases(self):
        return self.get_data(RepositoryItemType.RELEASES)


class OnlineRepositoryProvider(RepositoryProvider):
    """Provides direct access to the data from a GitHub repository."""

    def __init__(self, api):
        """Creates a new instance.
        
        :param api: a ghapi.core.GhApi object
        """

        self._api = api
        self._calls = {
            RepositoryItemType.COMMENTS: api.issues.list_comments_for_repo,
            RepositoryItemType.LABELS: api.issues.list_labels_for_repo,
            RepositoryItemType.EVENTS: api.issues.list_events_for_repo,
            RepositoryItemType.COMMITS: api.repos.list_commits,
            RepositoryItemType.RELEASES: api.repos.list_releases
            }
        self._calls_without_since = [api.issues.list_events_for_repo,
                                     api.repos.list_releases]

    def get_data(self, item_type, since=None):
        data = None
        if item_type == RepositoryItemType.ISSUES:
            data = self._fetch(self._api.issues.list_for_repo,
                               apiargs={'state': 'all'},
                               since=since)
        elif item_type == RepositoryItemType.TEAMS:
            data = self._fetch_teams()
        else:
            data = self._fetch(self._calls[item_type], since=since)
        return data

    def _fetch(self, apicall, apiargs={}, since=None):
        """Generic method to fetch data from GitHub."""

        if since is not None:
            if apicall in self._calls_without_since:
                # No support for 'since=' parameter in those calls,
                # we need to ensure manually that we don't get more
                # than what we want
                return self._fetch_since(apicall, since, apiargs)
            apiargs['since'] = date2gh(since)

        things = apicall(per_page=100, **apiargs)
        last_page = self._api.last_page()
        page = 0

        while page < last_page:
            page += 1
            things.extend(apicall(per_page=100, page=page, **apiargs))

        return things

    def _fetch_since(self, apicall, since, apiargs={}):
        """Specialized method for items without 'since=' support."""

        things = apicall(per_page=100, **apiargs)
        last_page = self._api.last_page()
        page = 0

        loop = True
        while loop and page < last_page:
            if gh2date(things[-1].created_at) <= since:
                # Stop fetching if we got what we were looking for
                loop = False
            else:
                page += 1
                things.extend(apicall(per_page=100, page=page, **apiargs))

        # Remove everything before the cutoff timestamp
        return [i for i in things if gh2date(i.created_at) >= since]

    def _fetch_teams(self):
        teams = [AttrDict({'slug': '__collaborators'})]
        try:
            teams.extend(self._fetch(self._api.teams.list))
        except HTTP4xxClientError:
            logging.warn("Cannot get teams list (possibly due to "
                         "insufficient access rights)")

        for team in teams:
            team['members'] = self._fetch_team_members(team.slug)

        return teams

    def _fetch_team_members(self, slug):
        members = []
        try:
            if slug == '__collaborators':
                members = self._fetch(self._api.repos.list_collaborators)
            else:
                members = self._fetch(self._api.teams.list_members_in_org,
                                      {'team_slug': slug})
        except HTTP4xxClientError:
            logging.warn("Cannot get team members (possibly due to "
                         "insufficient access rights)")

        return members


class MemoryRepositoryProvider(RepositoryProvider):
    """In-memory cache for data from a GitHub repository."""

    _wrappers = {
        RepositoryItemType.ISSUES: IssueItem,
        RepositoryItemType.COMMITS: CommitItem,
        RepositoryItemType.COMMENTS: RepositoryItem,
        RepositoryItemType.EVENTS: RepositoryItem,
        RepositoryItemType.RELEASES: RepositoryItem
        }

    def __init__(self, backend):
        """Creates a new instance.
        
        :param backend: a RepositoryProvider object from which to fetch
            the data that are not already in memory
        """

        self._backend = backend
        self._data = {}

    def get_data(self, item_type, since=None):
        if item_type not in self._data:
            items = self._backend.get_data(item_type, since)
            wrapper = self._wrappers.get(item_type, None)
            if wrapper is not None:
                for item in items:
                    item.__class__ = wrapper
            self._data[item_type] = items
        return self._data[item_type]



# -- General configuration ------------------------------------------------

source_suffix = '.rst'
master_doc = 'index'

copyright = u'2021,2022 Damien Goutte-Gattat'
author = u'Damien Goutte-Gattat <dpg44@cam.ac.uk>'

language = 'en'

exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

pygments_style = 'sphinx'

extensions = ['sphinx.ext.intersphinx']
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

# -- Options for HTML output ----------------------------------------------

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']

# -- Options for LaTeX output ---------------------------------------------

latex_engine = 'lualatex'

latex_elements = {
        'papersize': 'a4paper',
        'pointsize': '10pt'
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'GrainyHead.tex', u'GrainyHead Documentation',
     u'Damien Goutte-Gattat', 'manual'),
]

# -- Options for manual page output ---------------------------------------

man_pages = [
    (master_doc, 'grainyhead', u'GrainyHead Documentation',
     [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (master_doc, 'GrainyHead', u'GrainyHead Documentation',
     author, 'GrainyHead', 'Helper tools for GitHub.',
     'Miscellaneous'),
]



# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from configparser import ConfigParser, _UNSET
from time import sleep
from random import randint
import sys
import re
import os

import click
from click_shell import shell
from ghapi.core import GhApi

from . import __version__
from .repository import Repository
from .providers import OnlineRepositoryProvider
from .caching import FileRepositoryProvider
from .util import Date, Interval
from .metrics import MetricsFormatter, MetricsReporter

prog_name = "grh"
prog_notice = f"""\
{prog_name} (GrainyHead {__version__})
Copyright © 2022 Damien Goutte-Gattat

This program is released under the GNU General Public License.
See the COPYING file or <http://www.gnu.org/licenses/gpl.html>.
"""


def die(msg):
    print(f"{prog_name}: {msg}", file=sys.stderr)
    sys.exit(1)


def _parse_github_url(url):
    m = re.match('(https?://github.com/)?([^/]+)/([^/.]+)', url)
    if not m:
        die(f"Invalid GitHub repository: {url}.")

    return m.groups()[1:]


class GrhContext(object):

    def __init__(self, config_file, section='default'):
        self._config_file = config_file
        self._name = section

        self._config = ConfigParser()
        self._has_config = len(self._config.read(config_file)) > 0

        self._repo = None
        self._cache = True

    def reset(self, section='default', config_file=None, options=None):
        self._repo = None
        self._name = section

        if config_file:
            self._config.clear()
            self._config_file = config_file
            self._has_config = len(self._config.read(config_file)) > 0
        elif options:
            self._config.clear()
            self._config.add_section(section)
            for key, value in options.items():
                self._config.set(section, key, value)
            self._has_config = True

    def get_option(self, key, fallback=_UNSET):
        return self._config.get(self._name, key, fallback=fallback)

    def disable_cache(self):
        self._cache = False

    @property
    def repository(self):
        if not self._repo:
            repo_url = self._config.get(self._name, 'repository')
            owner, repo = _parse_github_url(repo_url)
            token = self._config.get(self._name, 'token', fallback=None)
            api = GhApi(owner=owner, repo=repo, org=owner, token=token)
            backend = OnlineRepositoryProvider(api)
            if self._cache:
                backend = FileRepositoryProvider(self.cache_dir, backend)
            self._repo = Repository(api, backend)
        return self._repo

    @property
    def has_config(self):
        return self._has_config

    @property
    def config_file(self):
        return self._config_file

    @property
    def config(self):
        return self._config

    @property
    def cache_dir(self):
        xdg_data_dir = os.getenv('XDG_DATA_HOME',
                                 os.path.join(os.getenv('HOME'),
                                              '.local', 'share'))
        return os.path.join(xdg_data_dir, 'grainyhead', self._name)


@shell(context_settings={'help_option_names': ['-h', '--help']},
       prompt="grh> ")
@click.option('--config', '-c', type=click.Path(exists=False),
              default='{}/config'.format(click.get_app_dir('grainyhead')),
              help="Path to an alternative configuration file.")
@click.option('--section', '-s', default='default',
              help="Name of the configuration file section to use.")
@click.option('--no-cache', is_flag=True, default=False,
              help="Disable the file cache.")
@click.version_option(version=__version__, message=prog_notice)
@click.pass_context
def grh(ctx, config, section, no_cache):
    """Command-line tool for GitHub."""

    context = GrhContext(config, section)
    if no_cache:
        context.disable_cache()
    ctx.obj = context
    if not context.has_config:
        ctx.invoke(conf)


@grh.command(name='issues')
@click.option('--older-than', 'cutoff', default='1y', type=Date,
              help="""Only list issues that have not been updated since the
                      specified date or for the specified duration
                      (default=1y, or one year ago).""")
@click.option('--team', default='__collaborators', metavar='NAME',
              help="""The name of a GitHub team.""")
@click.pass_obj
def list_issues(grh, cutoff, team):
    """List open issues.
    
    This commands list open issues that have not been updated for a
    given amount of time.
    """

    repo = grh.repository

    issues = [i for i in repo.issues if i.updated(before=cutoff)]
    members = [m.login for m in repo.get_team(team)]

    print("| Issue | Author | Team? | Assignee(s) |")
    print("| ----- | ------ | ----- | ----------- |")

    for issue in issues:
        assignees = ', '.join(['@[{}]({})'.format(a.login, a.html_url) for a in issue.assignees])
        is_team = 'Yes' if issue.user.login in members else 'No'

        print(f"| [{issue.title}]({issue.html_url}) | @[{issue.user.login}]({issue.user.html_url}) | {is_team} | {assignees} |")


@grh.command(name='close')
@click.option('--older-than', 'cutoff', default='1y', type=Date,
              help="""Close issues that have not been updated since the
                      specified date or for the specified duration
                      (default=1y, or one year ago).""")
@click.option('--comment', '-c', default=None,
              help="""Text of the comment to add to the closed issues.""")
@click.option('--dry-run', '-d', is_flag=True, default=False,
              help="""List issues that would be closed without actually
                      closing them.""")
@click.option('--limit', '-l', default=-1, metavar='N',
              help="""Only close the N oldest issues.""")
@click.pass_obj
def auto_close(grh, comment, cutoff, dry_run, limit):
    """Close old issues.
    
    This command automatically closes issues that have not been updated
    for a given amount of time.
    """

    repo = grh.repository

    repo.create_label('autoclosed-unfixed', 'ff7000',
                      'This issue has been closed automatically.')

    if not comment:
        comment = grh.get_option('close.comment', fallback=None)

    issues = [i for i in reversed(repo.issues) if i.updated(before=cutoff)]
    if limit != -1:
        issues = issues[:limit]

    click.echo_via_pager(_list_closable_issues(issues))
    if comment:
        click.echo(f"Closing with comment: {comment}")
    if dry_run or not click.confirm("Proceed?"):
        return

    with click.progressbar(issues, item_show_func=_show_closing_issue) as bar:
        for issue in bar:
            repo.close_issue(issue, 'autoclosed-unfixed', comment)

            # GitHub's documentation says that requests that trigger
            # notifications (such as adding a comment to an issue)
            # should be issued "at a reasonable pace", but does not
            # elaborate on what a "reasonable pace" is.
            # We try here to wait for anything between 2 and 10 seconds
            # between requests.
            sleep(randint(2, 10))


def _list_closable_issues(issues):
    yield f"The following {len(issues)} issues are about to be closed:\n"
    for issue in issues:
        last_update, _ = issue.updated_at.split('T')
        yield f"{issue.number}: last update {last_update}: {issue.title}\n"


def _show_closing_issue(issue):
    if issue:
        return f"Closing issue #{issue.number}"
    else:
        return ""


@grh.command()
@click.option('--from', 'start', type=Date, default='6m',
              help="""Set the beginning of the reporting period.
                      Default to 6m (6 months ago).""")
@click.option('--to', 'end', type=Date, default='now',
              help="""Set the end of the reporting period.
                      Default to now.""")
@click.option('--team', default='__collaborators', metavar='NAME',
              help="""The name of a GitHub team.""")
@click.option('--selector', '-s', multiple=True, default=[], metavar='SELECTOR',
              help="""Select a subset of repository events.""")
@click.option('--format', '-f', 'fmt', default='markdown',
              type=click.Choice(['json', 'markdown', 'csv', 'tsv']),
              help="""Write output in the specified format.""")
@click.option('--period', '-p', type=Interval, default=None,
              help="""Break down the metrics per periods of the
                      specified duration.""")
@click.pass_obj
def metrics(grh, start, end, team, selector, fmt, period):
    """Get repository metrics.
    
    This command collects and prints the number of repository events
    (such as the opening and closing of issues, commits, etc.) that
    occurred over a given period of time, as well as the number of
    individual contributors that caused those events.
    """

    reporter = MetricsReporter(grh.repository)
    if len(selector) == 0:
        selector = ['all AS Total',
                    f'team:{team} AS Internal',
                    f'!team:{team} AS External']

    metrics = reporter.get_report(selector, start, end, period)

    formatter = MetricsFormatter.get_formatter(fmt)
    formatter.write(metrics, sys.stdout)


@grh.command()
@click.pass_obj
def conf(grh):
    """Edit the configuration.
    
    This command either creates a new basic configuration file for use
    with a repository or starts a text editor to edit the current
    configuration file if it already exists.
    """

    if grh.has_config:
        click.termui.edit(filename=grh.config_file)
        grh.reset(config_file=grh.config_file)
    else:
        opts = {}
        opts['repository'] = click.termui.prompt("Repository name or URL")

        click.echo("Visit https://github.com/settings/tokens to create "
                   "a personal access token.")
        opts['token'] = click.termui.prompt("Token")
        grh.reset(options=opts)

        with open(grh.config_file, 'w') as f:
            grh.config.write(f)


def refresh(grh):
    pass


try:
    from IPython import embed

    @grh.command(name='ipython')
    @click.pass_obj
    def python_shell(grh):
        """Start an interactive Python shell.
    
        This commands starts a Python shell from where the GitHub API
        can be directly manipulated through the 'api' object. This is
        intended for testing purposes.
        """

        api = grh.repository._api
        repo = grh.repository
        embed()

except ImportError:
    pass

if __name__ == '__main__':
    grh()



# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import json
from datetime import timedelta


class MetricsReporter(object):
    """Generate metrics about events in a repository."""

    def __init__(self, repository):
        """Create a new instance.
        
        :param repository: the repository to work with
        """

        self._repo = repository

    def get_report(self, selectors, start, end, period=None):
        selectors = self._expand_wildcard_selectors(selectors)

        if period is None:
            return self._get_report_for_period(selectors, start, end)
        else:
            done = False
            reports = []

            while not done:
                period_end = start + period - timedelta(days=1)
                reports.append(self._get_report_for_period(selectors, start,
                                                           period_end))

                if period_end > end:
                    done = True
                else:
                    start = start + period

            return reports

    def _get_report_for_period(self, selectors, start, end):
        rset = _MetricsReportSet(start, end)
        self._date_filter = DateRangeFilter(start, end)

        for selector in selectors:
            item_filter = self._get_filter_from_selector(selector)
            rset.contributions.append(self.get_single_report(item_filter))

        return rset

    def get_single_report(self, item_filter):
        """Get a single report object based on the given filter."""

        issues_opened = [i for i in self._repo.all_issues
                         if item_filter.filterIssue(i)]
        issues_closes = [e for e in self._repo.events
                         if e.event == 'closed'
                         and not hasattr(e.issue, 'pull_request')
                         and item_filter.filterEvent(e)]

        pulls_opened = [p for p in self._repo.all_pull_requests
                        if item_filter.filterIssue(p)]
        pulls_closes = [e for e in self._repo.events
                        if e.event == 'closed'
                        and hasattr(e.issue, 'pull_request')
                        and item_filter.filterEvent(e)]
        pulls_merged = [e for e in self._repo.events
                        if e.event == 'merged'
                        and item_filter.filterEvent(e)]

        comments = [c for c in self._repo.comments
                    if item_filter.filterComment(c)]

        commits = [c for c in self._repo.commits
                   if item_filter.filterCommit(c)]

        releases = [r for r in self._repo.releases
                    if item_filter.filterRelease(r)]

        contributors = []
        contributors.extend([i.user.login for i in issues_opened])
        contributors.extend([p.user.login for p in pulls_opened])
        contributors.extend([c.user.login for c in comments])
        contributors.extend([e.actor.login for e in issues_closes])
        contributors.extend([e.actor.login for e in pulls_closes])
        contributors = set(contributors)

        return _Report(str(item_filter), item_filter.name, [
            len(issues_opened),
            len(issues_closes),
            len(pulls_opened),
            len(pulls_closes),
            len(pulls_merged),
            len(comments),
            len(commits),
            len(releases),
            len(contributors)])

    def _expand_wildcard_selectors(self, selectors):
        if not True in ['*' in s for s in selectors]:
            return selectors

        expanded_selectors = []
        for selector in selectors:
            if selector == 'user:*':
                expanded_selectors.extend([f'user:{u}' for u in self._repo.contributors])
            elif selector == 'label:*':
                expanded_selectors.extend([f'label:{l}' for l in self._repo.labels])
            else:
                expanded_selectors.append(selector)

        return expanded_selectors

    def _get_filter_from_selector(self, selector, level=0):
        if ' AS ' in selector:
            selector, name = selector.split(' AS ')
        else:
            name = selector

        if selector[0] == '!':
            f = ComplementFilter(self._get_filter_from_selector(selector[1:], level + 1))
        elif selector == 'all':
            f = NullFilter()
        elif selector.startswith('team:'):
            team_slug = selector[5:]
            members = [m.login for m in self._repo.get_team(team_slug)]
            f = TeamFilter(team_slug, members)
        elif selector.startswith('user:'):
            user_name = selector[5:]
            f = UserFilter(user_name)
            if selector == name:
                name = f'@{user_name}'
        elif selector.startswith('label:'):
            label = selector[6:]
            f = LabelFilter(label)
            if selector == name:
                name = label
        else:
            f = NullFilter()

        if level == 0:
            return CombinedFilter(name, [self._date_filter, f])
        else:
            return f


class MetricsFormatter(object):
    """Write a MetricsReportSet object."""

    def write(self, metrics, output):
        pass

    @staticmethod
    def get_formatter(fmt):
        if fmt.lower() == 'markdown':
            return MarkdownMetricsFormatter()
        elif fmt.lower() == 'csv':
            return CsvMetricsFormatter()
        elif fmt.lower() == 'tsv':
            return CsvMetricsFormatter(sep='\t')
        else:
            # We default to JSON
            return JsonMetricsFormatter()


class JsonMetricsFormatter(MetricsFormatter):

    def write(self, metrics, output):
        if isinstance(metrics, list):
            d = [self._get_dict_for_period(m) for m in metrics]
        else:
            d = self._get_dict_for_period(metrics)
        json.dump(d, output, indent=2)

    def _get_dict_for_period(self, metrics):
        d = {
            'period': {
                'to': metrics.end_date.strftime('%Y-%m-%d'),
                'from': metrics.start_date.strftime('%Y-%m-%d')
                },
            'contributions': []
            }

        for report in metrics.contributions:
            c = {
                'selector': report.selector,
                'results': {
                    'contributors': report.contributors,
                    'issues': {
                        'opened': report.issues_opened,
                        'closed': report.issues_closed
                        },
                    'pull_requests': {
                        'opened': report.pull_requests_opened,
                        'closed': report.pull_requests_closed,
                        'merged': report.pull_requests_merged
                        },
                    'comments': report.comments,
                    'commits': report.commits,
                    'releases': report.releases
                    }
                }
            d['contributions'].append(c)

        return d


class MarkdownMetricsFormatter(MetricsFormatter):

    def write(self, reportset, output):
        if isinstance(reportset, list):
            for r in reportset:
                self._write_reportset(r, output)
        else:
            self._write_reportset(reportset, output)

    def _write_reportset(self, reportset, output):
        start = reportset.start_date
        end = reportset.end_date

        with_total = reportset.contributions[0].selector == 'all'

        output.write(f"From {start:%Y-%m-%d} to {end:%Y-%m-%d}\n")
        output.write("\n")

        output.write("| Event                |")
        if with_total:
            output.write(f" {reportset.contributions[0].name:8.8} |")
            for report in reportset.contributions[1:]:
                output.write(f" {report.name:8.8} | {report.name:4.4} (%) |")
            output.write("\n")
            output.write("| -------------------- | -------- |")
            for report in reportset.contributions[1:]:
                output.write(" -------- | -------- |")
            output.write("\n")
        else:
            for report in reportset.contributions:
                output.write(f" {report.name:8.8} |")
            output.write("\n")
            output.write("| -------------------- |")
            for report in reportset.contributions:
                output.write(" -------- |")
            output.write("\n")

        items = [
            "Contributors",
            "Issues opened",
            "Issues closed",
            "Pull requests opened",
            "Pull requests closed",
            "Pull requests merged",
            "Comments",
            "Commits",
            "Releases"
            ]

        for item in items:
            self._write_line(item, reportset.contributions, output, with_total)
        output.write('\n')

    def _write_line(self, label, reports, output, with_total):
        property_name = label.lower().replace(' ', '_')

        if with_total:
            total = getattr(reports[0], property_name)
            output.write(f"| {label:20} | {total: 8} |")

            for report in reports[1:]:
                value = getattr(report, property_name)
                if total > 0:
                    percent = value / total * 100
                else:
                    percent = 0.0
                output.write(f" {value:8} | {percent: 8.2f} |")
        else:
            output.write(f"| {label:20} |")
            for report in reports:
                value = getattr(report, property_name)
                output.write(f" {value:8} |")
        output.write("\n")


class CsvMetricsFormatter(MetricsFormatter):

    def __init__(self, sep=','):
        self._sep = sep

    def write(self, reportset, output):
        headers = ['Date', 'Selector', 'Selector name',
                   'Issues opened', 'Issues closed',
                   'Pull requests opened', 'Pull requests closed',
                   'Pull requests merged', 'Comments', 'Commits',
                   'Releases', 'Contributors']
        output.write(self._sep.join(headers))
        output.write('\n')

        if isinstance(reportset, list):
            for r in reportset:
                self._write_reportset(r, output)
        else:
            self._write_reportset(reportset, output)

    def _write_reportset(self, reportset, output):
        for report in reportset.contributions:
            values = [reportset.end_date.strftime('%Y-%m-%d'),
                      report.selector, report.name,
                      report.issues_opened, report.issues_closed,
                      report.pull_requests_opened, report.pull_requests_closed,
                      report.pull_requests_merged, report.comments,
                      report.commits, report.releases, report.contributors]
            output.write(self._sep.join([str(v) for v in values]))
            output.write('\n')


class _MetricsReportSet(object):
    """This object holds some metrics about events in a repository."""

    def __init__(self, start, end):
        """Create a new instance for the specified reporting period."""

        self._start = start
        self._end = end
        self._reports = []

    @property
    def start_date(self):
        """The beginning of the reporting period."""

        return self._start

    @property
    def end_date(self):
        """The end of the reporting period."""

        return self._end

    @property
    def contributions(self):
        """The individual contribution reports."""

        return self._reports


class _Report(object):
    """This object holds metrics computed from a given selector."""

    def __init__(self, selector, name, values):
        self._selector = selector
        self._name = name
        self._values = values

    @property
    def selector(self):
        return self._selector

    @property
    def name(self):
        return self._name

    @property
    def contributors(self):
        return self._values[8]

    @property
    def issues_opened(self):
        return self._values[0]

    @property
    def issues_closed(self):
        return self._values[1]

    @property
    def pull_requests_opened(self):
        return self._values[2]

    @property
    def pull_requests_closed(self):
        return self._values[3]

    @property
    def pull_requests_merged(self):
        return self._values[4]

    @property
    def comments(self):
        return self._values[5]

    @property
    def commits(self):
        return self._values[6]

    @property
    def releases(self):
        return self._values[7]


class ItemFilter(object):

    def filterIssue(self, issue):
        return self._filterItem(issue)

    def filterEvent(self, event):
        return self._filterItem(event)

    def filterComment(self, comment):
        return self._filterItem(comment)

    def filterCommit(self, commit):
        return self._filterItem(commit)

    def filterRelease(self, release):
        return self._filterItem(release)

    def _filterItem(self, _):
        return False

    @property
    def name(self):
        return str(self)


class NullFilter(ItemFilter):

    def _filterItem(self, _):
        return True

    def __str__(self):
        return 'all'


class DateRangeFilter(ItemFilter):

    def __init__(self, start, end):
        self._start = start
        self._end = end

    def __str__(self):
        return f'date:{self._start:%Y-%m-%d}..{self._end:%Y-%m-%d}'

    def _filterItem(self, item):
        return item.created(after=self._start, before=self._end)


class TeamFilter(ItemFilter):

    def __init__(self, team_name, members):
        self._members = members
        self._slug = team_name

    def __str__(self):
        return f'team:{self._slug}'

    def filterIssue(self, issue):
        return issue.user.login in self._members

    def filterEvent(self, event):
        return event.actor.login in self._members

    def filterComment(self, comment):
        return comment.user.login in self._members

    def filterCommit(self, commit):
        if commit.author is not None:
            return commit.author.login in self._members
        else:
            return False

    def filterRelease(self, release):
        return release.author.login in self._members


class UserFilter(TeamFilter):

    def __init__(self, user_name):
        TeamFilter.__init__(self, None, [user_name])

    def __str__(self):
        return f'user:{self._members[0]}'


class LabelFilter(ItemFilter):

    def __init__(self, label):
        self._label = label

    def __str__(self):
        return f'label:{self._label}'

    def filterIssue(self, issue):
        return self._label in [l.name for l in issue.labels]

    def filterEvent(self, event):
        return self._label in [l.name for l in event.issue.labels]


class ComplementFilter(ItemFilter):

    def __init__(self, inner_filter):
        self._filter = inner_filter

    def __str__(self):
        return f'!{str(self._filter)}'

    def filterIssue(self, issue):
        return not self._filter.filterIssue(issue)

    def filterEvent(self, event):
        return not self._filter.filterEvent(event)

    def filterComment(self, comment):
        return not self._filter.filterComment(comment)

    def filterCommit(self, commit):
        return not self._filter.filterCommit(commit)

    def filterRelease(self, release):
        return not self._filter.filterRelease(release)


class CombinedFilter(ItemFilter):

    def __init__(self, name, filters):
        self._name = name
        self._filters = filters

    def __str__(self):
        components = [str(f) for f in self._filters if not str(f).startswith('date:')]
        return ' & '.join(components)

    @property
    def name(self):
        return self._name

    def filterIssue(self, issue):
        return not False in [f.filterIssue(issue) for f in self._filters]

    def filterEvent(self, event):
        return not False in [f.filterEvent(event) for f in self._filters]

    def filterComment(self, comment):
        return not False in [f.filterComment(comment) for f in self._filters]

    def filterCommit(self, commit):
        return not False in [f.filterCommit(commit) for f in self._filters]

    def filterRelease(self, release):
        return not False in [f.filterRelease(release) for f in self._filters]




# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from .providers import MemoryRepositoryProvider


class Repository(object):

    def __init__(self, api, backend):
        self._api = api
        self._provider = MemoryRepositoryProvider(backend)
        self._labels = None
        self._teams = None

    @property
    def issues(self):
        return [i for i in self._provider.issues if i.closed_at is None]

    @property
    def all_issues(self):
        return self._provider.issues

    @property
    def pull_requests(self):
        return [i for i in self._provider.pull_requests if i.closed_at is None]

    @property
    def all_pull_requests(self):
        return self._provider.pull_requests

    @property
    def comments(self):
        return self._provider.comments

    @property
    def events(self):
        return self._provider.events

    @property
    def commits(self):
        return self._provider.commits

    @property
    def releases(self):
        return self._provider.releases

    @property
    def labels(self):
        if self._labels is None:
            self._labels = [l.name for l in self._provider.labels]
        return self._labels

    @property
    def contributors(self):
        return [m.login for m in self.get_team()]

    def get_team(self, name='__collaborators'):
        if self._teams is None:
            self._teams = {}
            for team in self._provider.teams:
                self._teams[team.slug] = team
        return self._teams[name].members

    def create_label(self, name, color, description):
        if not name in self.labels:
            self._api.issues.create_label(name, color, description)
            self._labels.append(name)

    def close_issue(self, issue, label=None, comment=None):
        if label:
            self._api.issues.add_labels(issue.number, [label])
        if comment:
            self._api.issues.create_comment(issue.number, comment)
        self._api.issues.update(issue.number, state='closed')



# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta, timezone
from dateutil.relativedelta import relativedelta
import re

import click

_durations = { 'd': 1, 'w': 7, 'm': 30, 'y': 365 }
_periods = { 'd': 'days', 'w': 'weeks', 'm': 'months', 'y': 'years'}
_date_formats = [
    '%Y-%m-%d',
    '%Y-%m'
    ]


class DateParamType(click.ParamType):
    """A parameter type for Click representing a date.

    This differs from the standard click.DateTime in that it also allows
    to specify a date as a number of days, weeks, months, or years from
    the current date. It also recognizes some special values.

    Specifically, this parameter accepts:
    - a plain date, written as YYYY-MM-DD or YYYY-MM;
    - Xd (or simply X), for the date X day(s) ago;
    - Xw, for the date X week(s) ago;
    - Xm, for the date X month(s) ago;
    - Xy, for the date Y year(s) ago;
    - 'now', for the current date;
    - 'origin', for the earliest possible date.
    """

    name = 'date'

    def convert(self, value, param, ctx):
        if isinstance(value, datetime):
            return value

        if value.lower() == 'now':
            return datetime.now(timezone.utc)
        elif value.lower() == 'origin':
            return datetime.min.replace(tzinfo=timezone.utc)
        elif m := re.match('^([0-9]+)([dwmy])?$', value):
            n, f = m.groups()
            if not f:
                f = 'd'
            now = datetime.now(timezone.utc)
            delta = timedelta(days=int(n) * _durations[f])
            return now - delta
        else:
            for fmt in _date_formats:
                try:
                    dt = datetime.strptime(value, fmt)
                    return dt.replace(tzinfo=timezone.utc)
                except ValueError:
                    pass
            self.fail(f"Cannot convert '{value}' to a date", param, ctx)


class TimeIntervalParamType(click.ParamType):
    """A parameter type for Click representing a time interval.
    
    This parameter accepts:
    - Xd (or simply X), for an interval of X day(s);
    - Xw, for an interval of X week(s);
    - Xm, for an interval of X calendar month(s);
    - Xy, for an interval of X year(s);
    - 'weekly', as an alternative for '1w';
    - 'monthly', as an alternative for '1m';
    - 'quarterly', as an alternative for '3m';
    - 'yearly', as an alternative for '1y'.
    """

    name = 'interval'

    def convert(self, value, param, ctx):
        if isinstance(value, relativedelta):
            return value

        value = value.lower()
        if value == 'weekly':
            return relativedelta(weeks=1)
        elif value == 'monthly':
            return relativedelta(months=1)
        elif value == 'quarterly':
            return relativedelta(months=3)
        elif value == 'yearly':
            return relativedelta(years=1)
        elif m := re.match('^([0-9]+)([dwmy])?$', value):
            n, f = m.groups()
            if not f:
                f = 'd'
            d = {_periods[f]: int(n)}
            return relativedelta(**d)
        else:
            self.fail(f"Cannot convert '{value}' to a time interval", param, ctx)


Date = DateParamType()
Interval = TimeIntervalParamType()



# grainyhead - Helper tools for GitHub
# Copyright © 2021,2022 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup
from incenp.grainyhead import __version__

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='grainyhead',
    version=__version__,
    description='Helper tools for GitHub',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Damien Goutte-Gattat',
    author_email='dpg44@cam.ac.uk',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers'
        ],

    install_requires=[
        'click_shell >= 2.1',
        'ghapi',
        'python-dateutil'
        ],

    extras_require={
        'IPython': ['ipython']
        },

    packages=[
        'incenp',
        'incenp.grainyhead'
        ],

    entry_points={
        'console_scripts': [
            'grh = incenp.grainyhead.main:grh'
            ]
        },

    command_options={
        'build_sphinx': {
            'project': ('setup.py', 'GrainyHead'),
            'version': ('setup.py', __version__),
            'release': ('setup.py', __version__)
            }
        }
    )



# grainyhead - Helper tools for GitHub
# Copyright © 2021 Damien Goutte-Gattat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from datetime import datetime, timedelta
import json
import os.path
from os import makedirs
import time

from fastcore.xtras import dict2obj, obj2dict

from .providers import (RepositoryProvider, RepositoryItemType,
                        GITHUB_DATE_FORMAT)


class FileRepositoryProvider(RepositoryProvider):
    """Provides access to cached data from a GitHub repository."""

    def __init__(self, directory, backend):
        self._cachedir = directory
        self._backend = backend
        self._stale_cutoff = timedelta(days=30)

    def get_data(self, item_type, since=None):
        data = []
        now = time.time()
        refresh = False

        data_file = self._get_data_file(item_type)
        if os.path.exists(data_file):
            with open(data_file, 'r') as f:
                data = dict2obj(json.load(f))
            mtime = os.path.getmtime(data_file)
            if now - self._stale_cutoff.total_seconds() > mtime:
                refresh = True
                since = self._get_last_item_date(item_type, data)

        if len(data) == 0:
            refresh = True
            since = None

        if refresh:
            data.extend(self._backend.get_data(item_type, since))
            makedirs(self._cachedir, 0o755, True)
            with open(data_file, 'w') as f:
                json.dump(obj2dict(data), f, indent=0)

        return data

    def _get_data_file(self, item_type):
        filename = item_type.name.lower() + '.json'
        return os.path.join(self._cachedir, filename)

    def _get_last_item_date(self, item_type, data):
        if item_type == RepositoryItemType.ISSUES:
            # Force full refresh, so that we get updated status for
            # old issues
            return None
        elif item_type in [RepositoryItemType.LABELS, RepositoryItemType.TEAMS]:
            # Always fetch everything for those
            return None
        elif item_type == RepositoryItemType.COMMITS:
            # Only get new commits
            return datetime.strptime(data[-1].commit.author.date,
                                     GITHUB_DATE_FORMAT)
        else:
            # Only get new other items
            return datetime.strptime(data[-1].created_at, GITHUB_DATE_FORMAT)
